% GDB interface for Jed.  Derived from ashell.sl.
% Version 1.00  obscurity@o..., davis@s...
_debug_info = 1;
_traceback=1;
implements ("GDB");

% Customisation variables.
public variable GDB_Window_Mode = 1;% 0 = window soup
% 1 = keep it simple
public variable GDB_Secondary_Window_Max_Size = 7;  % only relevant when
% GDB_Window_Mode is 1
public variable GDB_Curr_Line_Mark_Style = "menu_selection"; % colour style
public variable GDB_Breakpoint_Mark_Style = "region";        % for line marks
public variable GDB_Cond_Breakpoint_Mark_Style = "status";   % for line marks
public variable GDB_Default_Data_Display_Mode = "hex";  % "hex" or "raw"
public variable GDB_Kill_Bufs_On_Exit = 1; % 0 = leave as-is,
% 1 = kill them
% 2 = bury them
public variable GDB_Trap_Quit = 1; % 0 = When 'quit' entered in *gdb* buffer,
%     pass it to gdb. Only use if debugging
%     an app that you need to send 'quit' to
%     - quitting at the prompt in the *gdb*
%     buffer causes gdb-mode to get confused
% 1 = Intercept 'quit' in *gdb* buffer and
%     treat it as though user has called
%     gdb_quit instead.
% End of customisation variables

!if (keymap_p ("GDB_keymap"))
   make_keymap ("GDB_keymap");
definekey("gdb_attach", "a", "GDB_keymap");
definekey("gdb_set_args", "A", "GDB_keymap");
definekey("gdb_break", "b", "GDB_keymap");
definekey("gdb_break_with_condition", "B", "GDB_keymap");
definekey("gdb_continue", "c", "GDB_keymap");
definekey("gdb_detach", "d", "GDB_keymap");
definekey("gdb_finish", "f", "GDB_keymap");
definekey("gdb_jump", "j", "GDB_keymap");
definekey("gdb_kill", "k", "GDB_keymap");
definekey("gdb_next", " ", "GDB_keymap");
definekey("gdb_run", "r", "GDB_keymap");
definekey("gdb_step", "\r", "GDB_keymap");
definekey("gdb_other", "o", "GDB_keymap");
definekey("gdb_quit", "q", "GDB_keymap");
definekey ("gdb_up", "^?", "GDB_keymap");
definekey ("gdb_output_identifier_at_cursor", "?", "GDB_keymap");
definekey ("gdb_watch_identifier_at_cursor", "w", "GDB_keymap");
definekey ("gdb_delete_watch", "W", "GDB_keymap");
definekey ("gdb_show_output_buffer","1", "GDB_keymap");
definekey("gdb_show_data_buffer","2", "GDB_keymap");
definekey("gdb_show_watch_buffer", "3", "GDB_keymap");
%definekey("gdb_show_debug_buffer","4", "GDB_keymap");
definekey("gdb_show_source_buffer_only","0", "GDB_keymap");
definekey("gdb_toggle_data_display_mode","t", "GDB_keymap");
definekey("gdb_edit_source","e","GDB_keymap");
definekey("gdb_load_source","l","GDB_keymap");
definekey ("gdb_run_to_current_line", "z","GDB_keymap");

public variable GDB_Output;

static variable Data_Display_Mode;
static variable Temp_Raw_Mode;

static variable Object;

static variable Source_Buffers;
static variable Pid = -1; static variable Args = ""; static variable Lineno;

static variable Origbuf;
static variable Curr_Sourcebuf; static variable Curr_Sourcefile;

static variable Process_Output_Hook;

static variable Identifier_Name;
static variable Identifier_Type; static variable Identifier_Size;

static variable Curr_Line_Mark;

static variable Breakpoints;
static variable Last_Breakpoint_Mark;

static variable Msg_Text;
static variable Msg_Priority;

static variable Prompt_Mark;

%() = evalfile("textmac.sl");     % defines filename_extension, which we use

%define debug(str)
%{
%   variable buf;
%   buf = whatbuf();
%   setbuf("*gdb-debug*");
%   eob;
%   insert("Debug: " + str + "\n");
%   setbuf(buf);
%}

% Each command sent to gdb may generate more than one valid response to
% the user - eg, "Continuing..." followed by "Program exited normally."
% The following 2 functions are used to determine which message to
% display - each message has a priority: 0 for info, 1 for warning etc.
define set_message(str, priority)
{
        if (priority > Msg_Priority)
        {
                Msg_Text = str;
                Msg_Priority = priority;
        }
}
define show_message()
{
        message(Msg_Text);
        Msg_Text = "";
        Msg_Priority = -1;
}

define send_input (str)
{
        str = str + "\n";
        %  debug("!" + cmd);
        send_process (Pid, str);
        setbuf("*gdb*");
        eob();
        insert(str);
        GDB_Output = "";
        eob();
        eol();
}

define wait_for_gdb()
{
        while (eob(), bskip_chars(" \t\n"), not blooking_at ("(gdb)"))
        {
                get_process_input(30);
                update(1);
        }
}

define issue_command(cmd, func)
{
        Process_Output_Hook = func;
        send_input(cmd);
        wait_for_gdb();
        Process_Output_Hook = NULL;
}

define show_secondary_window(buf)
{
        variable lineno, windowsize;

        if (GDB_Window_Mode)  % Find out what where cursor is in window..
        {
                lineno = window_line;
                lineno--;
                go_up(lineno);        % ...and move it to top, so next window always
        }                                                % created at the botton.

        pop2buf(buf);

        if (GDB_Window_Mode)
        {
                % Resize window to either size of buffer, or max size (whicever is smaller)
                eob();
                windowsize = what_line();
                windowsize--;
                if (windowsize > GDB_Secondary_Window_Max_Size)
                {
                        windowsize = GDB_Secondary_Window_Max_Size;
                }
                if (windowsize < 2) % JED doesn't want to display 1-line windows :(
                {
                        windowsize = 2;
                }
                while(window_info('r') > windowsize)
                {
                        otherwindow();
                        enlargewin();
                        otherwindow();
                }
                while(window_info('r') < windowsize)
                {
                        enlargewin();
                }
                % Bugger about with buffer until bottom line of buffer is displayed in bottom
                % line of window. Yes, 'bugger about' *is* a technical term, why do you ask? :)
                eob();
                go_up(windowsize);
                bol();
                loop(windowsize-1)
                {
                        update(1);
                        go_down(1);
                }
                if (buf == "*gdb*")             %Show (gdb) prompt in gdb window.
                {
                        update(1);
                        go_down(1);
                        eol();
                        Prompt_Mark = create_user_mark();
                }
        }
        otherwindow();
        if (GDB_Window_Mode)
        {
                go_down(lineno);                % Move cursor back to where it was.
        }
}

public define gdb_show_data_buffer()
{
        show_secondary_window("*gdb-data*");
}
%public define gdb_show_debug_buffer()
%{
%   show_secondary_window("*gdb-debug*");
%}
public define gdb_show_output_buffer()
{
        show_secondary_window("*gdb*");
}
public define gdb_show_watch_buffer()
{
        show_secondary_window("*gdb-watch*");
}

static define ungdb_source_buffers ()
{
        variable s;

        %   debug("ungdb");

        foreach (Source_Buffers) using ("next")
        {
                s = ();
                !if (bufferp (s.name))
                        continue;
                setbuf (s.name);
		use_syntax_table("C");
                use_keymap (s.keymap);
                set_readonly (s.readonly);
        }
        Source_Buffers = NULL;
        Curr_Sourcebuf = NULL;
        Curr_Sourcefile = NULL;
}

static define gdb_add_source_buffer (buf)
{
        variable s;
        foreach (Source_Buffers) using ("next")
        {
                s = ();
                if (s.name == buf)
                        return;
        }

        variable flags = 0;
        (,,,flags) = getbuf_info(whatbuf());

        s = struct
        {
                name, keymap, readonly, next
        };
        s.name = buf;
        s.keymap = what_keymap ();
        s.readonly = flags & 0x08;
        s.next = Source_Buffers;
        Source_Buffers = s;
}

static define kill_buf(buf)
{
        variable flags;

        if (bufferp(buf))
        {
                setbuf(buf);
                (,,,flags) = getbuf_info();
                if (flags mod 2)  % Buffer modified, so tell jed to kill it
                {
                        ungetkey('k');
                }
                delbuf(buf);
        }
}

static define gdb_cleanup ()
{
        variable buf;

        ungdb_source_buffers ();

        Breakpoints = NULL;
        Curr_Line_Mark = NULL;

        loop (buffer_list())
        {
                buf = ();
                if (buf[[:3]] == "*gdb")
                {
                        if (GDB_Kill_Bufs_On_Exit == 1)
                        {
                                kill_buf(buf);
                        }
                        if (GDB_Kill_Bufs_On_Exit == 2)
                        {
                                bury_buffer(buf);
                        }
                }
        }
        if (GDB_Window_Mode)
        {
                pop2buf(Origbuf);
                onewindow();
        }
        flush ("gdb-mode terminated.");
}

define get_identifier()
{
        variable chars_allowed = "a-zA-Z0-9_[]";
        variable identifier_name = "<unknown>";
        variable o,p,;

        % If user has defined region, use that as identifier.
        if (is_visible_mark())
        {
                check_region();
                return bufsubstr();
        }

        push_spot();
        bskip_chars(chars_allowed);
        push_mark();
        p = _get_point;
        skip_chars(chars_allowed);
        if (_get_point != p)
        {
                identifier_name = bufsubstr();
        }
        pop_mark_0 ();
        pop_spot();
        return identifier_name;
}

public define gdb_show_source_buffer_only()
{
        pop2buf(Curr_Sourcebuf);
        if (GDB_Window_Mode)
        {
                onewindow();
        }
}

static define signal_handler (pid, flags, status)
{
        message ("GDB " + aprocess_stringify_status (pid, flags, status));
        gdb_cleanup ();
        Pid = -1;
}

define insert_data_raw(str)
{
        variable pos;
        variable lines;

        pos = is_substr(str,"(gdb)");
        str = str[[:pos-3]];

        lines = strlen(str)/75;
        if ((strlen(str) mod 75) > 0)
        {
                lines++;
        }
        pos = 1;
        loop(lines)
        {
                insert(substr(str,pos,75));
                insert("\n");
                pos += 75;
        }
        insert(substr(str,pos,-1));
}

define insert_data_hex(str)
{
        variable pos,pos2;
        variable lines;
        variable hexstr = "";
        variable ascstr = "";
        variable asc;

        %   debug("show data hex:" + str);
        flush("Formatting data...");

        str = "\n" + str;

        pos = is_substr(str,"\n");
        pos2 = is_substr(str,":");
        do
        {
                str = substr(str,1,pos-1) + substr(str,pos2+1,-1);;
                pos = is_substr(str,"\n");
                pos2 = is_substr(str,":");
        }
        while (pos2 > 0);

        str = strup(str_delete_chars(str, "\tx"));
        %  str now contains a string representing the contents of the identifier in
        %  hex, but with a 0 before each 2-digit byte. We'll just ignore the 0 later.

        pos = 2;   % skip first '0'
        pos2 = 1;
        loop(integer(Identifier_Size))
        {
                hexstr = hexstr + substr(str,pos,2);
                if ((pos2 mod 2) == 0)
                {
                        hexstr = hexstr + " ";
                }
                if ((pos2 mod 8) == 0)
                {
                        hexstr = hexstr + " ";
                }
                pos2++;
                () = sscanf(substr(str,pos,2), "%x", &asc);
                if ((asc > 31) and (asc < 127))
                {
                        ascstr = ascstr + char(asc);
                }
                else
                {
                        ascstr = ascstr + ".";
                }
                pos += 3;  % skip this byte and next '0' (see, told you!)
        }

        lines = integer(Identifier_Size)/16;
        if ((integer(Identifier_Size) mod 16) > 0)
        {
                lines++;
        }

        pos = 1;
        pos2 = 1;
        loop(lines)
        {
                insert(sprintf("%08x:   ",pos2 - 1));
                insert(substr(hexstr,pos,42));
                goto_column(56);
                insert("| " + substr(ascstr,pos2,16));
                goto_column(75);
                insert("|\n");
                pos += 42;
                pos2 += 16;
        }
}

define show_data(str)
{
        setbuf("*gdb-data*");
        erase_buffer();
        insert ("Identifier: " + Identifier_Name + "\n");
        insert ("Size: " + Identifier_Size + "\n");
        insert ("Type: " + Identifier_Type + "\n");

        if  (Data_Display_Mode == "raw")
        {
                insert_data_raw(GDB_Output);
                if (Temp_Raw_Mode) % OK, now set it back to hex
                {
                        Data_Display_Mode = "hex";
                        Temp_Raw_Mode = 0;
                }
        }
        else
        {
                insert_data_hex(GDB_Output);
        }

        bob();
        setbuf(Curr_Sourcebuf);
        gdb_show_data_buffer();
        flush(Identifier_Name + " displayed in " + Data_Display_Mode + " format.");
}

define update_curr_line_mark()
{
        variable buf;

        buf = whatbuf();
        setbuf(Curr_Sourcebuf);
        Curr_Line_Mark = create_line_mark(color_number(GDB_Curr_Line_Mark_Style));
        setbuf(buf); }

define show_source_buffer (file, linenum, withmark)
{
        % Obvious solution here would be to use find_file(file).
        % However, if the user happens to have a buffer named <file> open, we don't
        % want to use it, 'cos it might be a different version of the source, or
        % even a completely different file with the same name. Therefore, prefix
        % our version with '*gdb-' and always use that instead.
        variable exists;
        variable ext;

        ext = path_extname(file);
	ext = ext[[1:]];

        Curr_Sourcebuf = "*gdb-" + extract_filename(file) + "*";
        exists = bufferp(Curr_Sourcebuf);

        if (GDB_Window_Mode)
        {
                sw2buf(Curr_Sourcebuf);
        }
        else
        {
                pop2buf(Curr_Sourcebuf);
        }

        !if (exists)         % Didn't exist before, so put source code in it.
        {
                insert_file(file);
                create_blocal_var("srcfile");
                set_blocal_var(file,"srcfile");
        }

        gdb_add_source_buffer(Curr_Sourcebuf);
        goto_line (linenum);
	bol; skip_white();
        set_readonly(1);
        mode_hook(ext);
        use_keymap ("GDB_keymap");
        if (withmark)
                Curr_Line_Mark = create_line_mark(color_number(GDB_Curr_Line_Mark_Style));
        update (1);
}

define get_line_number(str)
{
        variable file, line;

        file = extract_element (str, 0, ':');
        line = extract_element (str, 1, ':');

        if ((file == NULL) or (line == NULL))
                return;

        Curr_Sourcefile = file[[2:]];
        show_source_buffer (file[[2:]], integer (line), 1); }

public define delete_watch(watchno)
{
        bob();
        if (bol_fsearch(sprintf("%03u",watchno)))    % Does watch already exist?
        {
                push_mark();                                    % If so, delete old value
                go_down(1);
                del_region();
                return(1);
        }
        else
        {
                return(0);
        }
}

define process_watch_output(line)
{
        variable pos, pos2;
        variable watchno, watchname, watchvalue;

        %   debug("watch:" + line);

        pos = is_substr(line,":");
        pos -= 2;
        watchno = integer(line[[:pos]]);
        pos = is_substr(line,",");
        pos++;
        pos2 = is_substr(line,"=");
        pos2 -= 3;
        watchname = line[[pos:pos2]];
        watchvalue = extract_element(line,1,'=');

        setbuf("*gdb-watch*");
        set_readonly(0);

        !if (delete_watch(watchno))  % Delete old version, if it exists
        {
                eob();          % Watch wasn't found, so add new one at end of buffer
        }

        insert(sprintf("%03u[%s]:%s\n",watchno,watchname,watchvalue));
        set_readonly(1);
}

% This function handles processing that needs to be done for every line
% of output from gdb (watches, change of line number/source file, etc.)
define generic_line_at_a_time(line)
{
        if ("Program exited " == substr(line, 1, 15))
        {
                set_message(line,9);
        }
        if ("Program terminated with signal " == substr(line,1,31))
        {
                set_message(line,10);
        }
        if ("warning: Source file is more recent than executable" == substr(line,1,51))
        {
                set_message("Warning! Source file more recent than executable!",10);
        }

        if ("\032\032" == substr (line, 1, 2))
        {
                get_line_number (line);
        }
        if (is_substr(line,"!Watch!") )
        {
                process_watch_output(line);
        }
}

% This function takes the output from gdb and calls func for each line
define handle_line_at_a_time(str, func)
{
        variable pos;

        while (pos = is_substr (str, "\n(gdb) "), pos)
        {
                pos--;
                variable line = str[[0:pos]];
                str = str[[pos+7:]];
                if (pos)
                {
                        foreach (strtok (line, "\n"))
                        {
                                line = ();
                                @func(line);
                        }
                }

                % Sometimes gdb will output several prompts in row with no newline
                if (0 == strncmp (str, "(gdb) ", 6))
                        str = "\n" + str;
        }
}

% 'dummy' function because when evaluating get_value_hook for the first time
% it doesn't know it's own address, so make it call this instead...
define get_value_hook2(str)
{
        show_data(GDB_Output);
}

define get_value_hook(str)
{
        if ((is_substr(str,"Cannot access memory at address"))
            or (is_substr(str,"which is in a register.")))
                % So much for hex mode - use raw mode instead...
        {
                Temp_Raw_Mode = 1;
                Data_Display_Mode = "raw";
                issue_command("print " + Identifier_Name, &get_value_hook2);
                return;
        }
        show_data(GDB_Output);
}

define get_size_line_hook(line)
{
        variable pos, cmd;

        pos = is_substr(line,"=");
        if (pos)
        {
                Identifier_Size = line[[pos+1:]];
                if (Data_Display_Mode == "hex")
                {
                        cmd = "x/" + strtrim(Identifier_Size) + "xb &" + Identifier_Name;
                }
                else
                {
                        cmd = "print " + Identifier_Name;
                }
                issue_command(cmd, &get_value_hook);
        }
        else
        {
                error("Unexpected output from gdb.");
        }
}
define get_size_hook(str)
{
        handle_line_at_a_time(str, &get_size_line_hook);
}

define get_type_hook(str)
{
        variable cmd;
        variable pos;

        if (is_substr(str,"parse error")
            or (is_substr(str,"nvalid")))
        {
                set_message("Parse error - invalid identifier.",5);
                return;
        }


        pos = is_substr(str,"(gdb)");
        if (pos)
        {
                Identifier_Type =str[[7:pos-3]];
        }
        else
        {
                Identifier_Type = str[[7:]];
        }
        cmd = "print sizeof(" + Identifier_Name + ")";
        issue_command(cmd, &get_size_hook);
}

define output_value()
{
        issue_command("ptype " + Identifier_Name, &get_type_hook);
}

public define gdb_toggle_data_display_mode()
{
        if ( Data_Display_Mode == "hex")
        {
                Data_Display_Mode = "raw";
        }
        else
        {
                Data_Display_Mode = "hex";
        }
        message("Data will be displayed in " + Data_Display_Mode + " format");
        if (buffer_visible("*gdb-data*"))
        {
                output_value();
        }
}

public define gdb_output_identifier_at_cursor ()
{

        Identifier_Size = "<unknown>";
        Identifier_Name = get_identifier();
        if (Identifier_Name == "<unknown>")
        {
                message("Cursor is not over a valid identifier.");
        }
        else
        {
                output_value();
        }
}

define info_source_line_hook(line)
{
        if (0 == strncmp ("Located in ", line, 11))
        {
                Curr_Sourcefile = line[[11:]];
                show_source_buffer (Curr_Sourcefile, 0, 0);
                if (GDB_Window_Mode)
                {
                        onewindow();
                }
        }
        else
        {
                if (0 == strncmp (line, "No Current Source File", 22))
                {
                        Curr_Sourcebuf = -1;
                }
        }
}

define info_source_hook(str)
{
        handle_line_at_a_time(str, &info_source_line_hook);
}

define other_command_hook(str)
{
        gdb_show_output_buffer();    % Show response from gdb to user command.
}

define attach_hook(str)
{
        if (is_substr(str,"A program is being debugged already."))
        {
                set_message("Already attached to a process. Detach it first.",1);
                issue_command("n", NULL);       %Tell gdb not to kill process
        }
        else
        {
                set_message("Process attached.",0);
        }
}

define detach_hook(str)
{
        if (is_substr(str,"Detaching from program:"))
        {
                set_message("Detached.",0);
        }
}

define continue_hook(str)
{
        if (is_substr(str,"The program is not being run."))
        {
                set_message("Can't continue - program isn't running.",1);
        }
        else
        {
                set_message("Continuing...",0);
        }
}

define unset_last_breakpoint()
{
        variable prev;
        variable breakpoint;

        prev = NULL;
        foreach (Breakpoints) using ("next")
        {
                breakpoint = ();
                if (breakpoint.mark == Last_Breakpoint_Mark)
                {
                        if (prev == NULL)
                                Breakpoints = breakpoint.next;
                        else
                                prev.next = breakpoint.next;

                        return;
                }
                else
                {
                        prev = breakpoint;
                }
        }
}

define break_hook(str)
{
        if (is_substr(str,"No source file named "))
        {
                set_message("Breakpoint not set!  Is current source part of program being debugged?",5);
                unset_last_breakpoint();
        }

        if (is_substr(str,"Cannot insert breakpoint"))
        {
                set_message("Breakpoint not set!  See gdb output for more details.",5);
                unset_last_breakpoint();
        }

        if (is_substr(str,"No symbol"))
        {
                set_message("Conditional brakpoint not set!  Did you enter the condition correctly?",5);
                unset_last_breakpoint();
        }
        set_message("Breakpoint set.",0);
}

define clear_hook(str)
{
        if (is_substr(str,"No breakpoint at"))
        {
                set_message("Breakpoint not found! (bug in gdb-mode - please report it)",5);
        }
        else
        {
                set_message("Breakpoint deleted.",0);
        }
}

define finish_line_hook(line)
{
        variable pos;

        pos = is_substr(line,"Value returned is $");
        if (pos)
        {
                pos=  is_substr(line," = ");
                pos += 3;
                set_message("Returned: " + line[[pos:]],0);
        }
}
define finish_hook(str)
{
        if (is_substr(str,"The program is not running"))
        {
                set_message("Can't finish - program not running.",1);
        }
        if (is_substr(str,"No selected frame."))
        {
                set_message("Can't finish - no frame selected (bug in gdb-mode?)",2);
        }
        if (is_substr(str,"not meaningful in the outermost frame."))
        {
                set_message("Can't finish - in top frame.",1);
        }
        if (is_substr(str,"Value returned is $"))
        {
                handle_line_at_a_time(str, &finish_line_hook);
        }
}

define jump_hook(str)
{
        if (is_substr(str,"The program is not being run"))
        {
                set_message("Can't jump - program not running.",0);
        }
        if (is_substr(str,"Unreasonable jump request"))
        {
                set_message("Can't jump - unreasonable request (bug in gdb-mode?)",2);
        }
        if (is_substr(str,"No source file has been specified."))
        {
                set_message("Can't jump - no source file.",2);
        }
        if (is_substr(str," is not in "))
        {
                if (get_y_or_n("Really jump out of function"))
                {
                        issue_command("y",NULL);
                        set_message("Jumped out of function.  Good luck!",1);
                }
                else
                {
                        issue_command("n",NULL);
                        set_message("Jump aborted.",1);
                }
        }
        if (is_substr(str,"WARNING!!!  Destination is in unmapped overlay!"))
        {
                if (get_yes_no("Destination in unmapped overlay - do you really know what you're doing"))
                {
                        set_message("Jumped, um, somewhere.  Just don't blame me...",1);
                        issue_command("y",NULL);
                        return;
                }
                else
                {
                        set_message("Jump aborted.",1);
                        issue_command("n",NULL);
                        return;
                }
        }
        if (is_substr(str,"Continuing at "))
        {
                update_curr_line_mark();
                set_message("Jumped.",0);
        }
}

define kill_hook(str)
{
        if (is_substr(str,"Kill the program being debugged?"))
        {
                if (get_y_or_n("Really kill program"))
                {
                        set_message("Program killed.",1);
                        issue_command("y",NULL);
                }
                else
                {
                        set_message("The program breathes a sigh of relief :)",1);
                        issue_command("n",NULL);
                }
        }
        if (is_substr(str,"The program is not being run."))
        {
                set_message("Can't kill - program not running.",1);
        }
        set_message("Program killed.",0);
}

define step_hook(str)
{
        if (is_substr(str,"The program is not being run."))
        {
                set_message("Can't step - program not running.",1);
        }
        if (is_substr(str,"No current frame"))
        {
                set_message("Can't step - no current frame (bug in gdb?)",2);
        }
}

define run_hook(str)
{
        if (is_substr(str,"The program being debugged has been started already."))
        {
                if (get_y_or_n("Program already running - restart"))
                {
                        set_message("Program restarted.",1);
                        issue_command("y",NULL);
                }
                else
                {
                        set_message("Restart aborted.",1);
                        issue_command("n",NULL);
                }
        }
        set_message("Program started.",0);
}

define run_to_current_line_hook(str)
{

        if (is_substr(str,"The program being debugged has been started already."))
        {
                issue_command("n",NULL);
                issue_command("continue", &continue_hook);
        }
}

define up_hook(str)
{
        if (is_substr(str,"No stack."))
        {
                set_message("Can't go up - no stack (is program running?).",1);
        }
        if (is_substr(str,"Initial frame selected; you cannot go up."))
        {
                set_message("Can't go up - in top frame.",1);
        }
        set_message("Previous stack frame selected.",0);
}

define watch_hook(str)
{
        if (is_substr(str,"No symbol "))
        {
                set_message("Can't add watch - symbol not defined (is program running?).",1);
        }
        if (is_substr(str,"optimized out"))
        {
                set_message("Can't add watch - symbol was optimized out.",1);
        }
        if (is_substr(str,"!Watch!"))
        {
                set_message("Watch added.", 0);
                gdb_show_watch_buffer();
        }
}

define user_command_hook()
{
        otherwindow();

        % This is a dirty little hack. For some reason that I can't figure, the
        % point is being moved 1 character to the left sometime after handle_output
        % exits (which should be the last thing processed!). This moves it before the
        % user mark defined for the start of input, and things go wrong.
        % As the only thing left to do after this function is call show_message(), I'm
        % doing that here and then triggering an error (with an empty message so we
        % don't alarm the user) which causes execution to halt.
        % This prevents the cursor being moved, and things work OK again :)
        % This is truly ugly, but the only way I can get things to work.  If anyone
        % can tell me why the cursor moves to the left without this, I'd be very
        % grateful.
        show_message();
        error(" ");
}

public define gdb_attach ()
{
        variable object_pid;

        object_pid = read_mini("Process ID to attach to : ","","");
        if (object_pid != "")
        {
                if (_slang_guess_type(object_pid) != Integer_Type)
                {
                        error("Process ID must be numeric.");
                        return;
                }
                if (kill(integer(object_pid), 0) == -1)
                {
                        verror("Process %s doesn't exist.", object_pid);
                        return;
                }
                issue_command("attach " + object_pid, &attach_hook);
        }
}

public define gdb_set_args ()
{
        Args = read_mini("Command line arguments for program: ","","");
}

public define gdb_detach ()
{
        issue_command("detach", &detach_hook);
}

define set_breakpoint(breaktype, location, condition)
{
        variable file;

        % user may have changed buffer on us, so use buffer-local variable to
        % determine sourcefile that breakpoint should be set in.
        (, file) = parse_filename(get_blocal_var("srcfile"));

        if (file == Curr_Sourcefile)
        {
                %GDB sometimes gets confused if you specify same file as currently
                %being used, so don't specify file if that's the case.
                issue_command(breaktype + " " + location + " " + condition, &break_hook);
        }
        else
        {
                issue_command(breaktype + " " + file + ":" + location, &break_hook);
        }
}

define gdb_toggle_break (withcond)
{
        variable breakpoint;
        variable prev;
        variable cond;

        push_spot ();
        bol ();
        Last_Breakpoint_Mark = create_user_mark ();
        pop_spot ();

        prev = NULL;
        foreach (Breakpoints) using ("next")
        {
                breakpoint = ();
                if (breakpoint.mark == Last_Breakpoint_Mark)
                {
                        if (prev == NULL)
                                Breakpoints = breakpoint.next;
                        else
                                prev.next = breakpoint.next;

                        issue_command("clear " +  breakpoint.location, &clear_hook);
                        return;
                }
                else
                {
                        prev = breakpoint;
                }
        }

        if (withcond)
        {
                cond = "if " + read_mini("Condition:","","");
        }
        else
        {
                cond = "";
        }

        breakpoint = struct {mark, location, linemark, next};
        breakpoint.mark = Last_Breakpoint_Mark;
        breakpoint.location = string (what_line ());
        if (withcond)
        {
                breakpoint.linemark = create_line_mark (color_number(GDB_Cond_Breakpoint_Mark_Style));
        }
        else
        {
                breakpoint.linemark = create_line_mark (color_number(GDB_Breakpoint_Mark_Style));
        }

        breakpoint.next = Breakpoints;
        Breakpoints = breakpoint;

        set_breakpoint("break", breakpoint.location, cond);
}

public define gdb_break()
{
        gdb_toggle_break(0);
}

public define gdb_break_with_condition()
{
        gdb_toggle_break(1);
}

public define gdb_continue()
{
        issue_command("continue", &continue_hook);
}
public define gdb_finish()
{
        issue_command("finish", &finish_hook);
}

public define gdb_jump()
{
        issue_command("jump " + string(what_line), &jump_hook);
}

public define gdb_kill()
{
        issue_command("kill", &kill_hook);
}

public define gdb_next()
{
        issue_command("next", &step_hook);
}

public define gdb_run()
{
        issue_command("run " + Args, &run_hook);
}

public define gdb_run_to_current_line()
{
        variable location;

        location = string(what_line());
        set_breakpoint("tbreak", location,"");
        issue_command("run " + Args, &run_to_current_line_hook);
}

public define gdb_quit ()
{
        flush ("Quitting gdb-mode....");
        send_input("quit");
        send_input("y");
        gdb_cleanup ();
}

public define gdb_step()
{
        issue_command("step", &step_hook);
}

public define gdb_up()
{
        issue_command("up", &up_hook);
}

public define gdb_other()
{
        variable str;
        str = read_mini("GDB>","","");
        issue_command(str, &other_command_hook);
}

public define gdb_watch_identifier_at_cursor ()
{
        variable id;

        id = get_identifier();
        if (id == "<unknown>")
        {
                message("Cursor is not over a valid identifier.");
        }
        else
        {
                issue_command("display \"!Watch!\", " + id, &watch_hook);
        }
}

public define gdb_delete_watch()
{
        variable watchno;
        variable deleted;

        watchno = read_mini("Enter watch number to be deleted>","","");

        if (_slang_guess_type(watchno) != Integer_Type)
        {
                error("Watch number must be numeric.");
        }
        else
        {
                setbuf("*gdb-watch*");
                set_readonly(0);
                deleted = delete_watch(integer(watchno));
                set_readonly(1);
                if (deleted)
                {
                        issue_command("undisplay " + watchno, NULL);
                        message("Watch deleted.");
                        gdb_show_watch_buffer();
                }
                else
                {
                        message("Watch number " + watchno + " not found.");
                }
        }
}

public define gdb_edit_source()
{
        variable buf = whatbuf();
        variable line = what_line;
        variable col = what_column;

        buf = substr(buf,6,strlen(buf)-6);
        if (bufferp(buf))
        {
                sw2buf(buf);
        }
        else
        {
                find_file(Curr_Sourcefile);
        }
        goto_line(line);
        goto_column(col);

        message("Editing " + buf);
}

public define gdb_load_source()
{
        variable file;

        file = read_with_completion("File to load: ","","",'f');
        if (stat_file (file) == NULL)
        {
                error ("Can't read " + file);
        }
        show_source_buffer (file, 0, 0);
}

public define gdb_process_buffer_input()
{
        variable user_input;
        variable quit_key;
        variable count;

        % Get user input...
        push_mark();
        goto_user_mark(Prompt_Mark);
        check_region(0);
        user_input = bufsubstr();
        % And now delete it, 'cos send_input will insert it and we don't want it twice
        push_mark();                                         % Have to do this again 'cos bufsubstr
        goto_user_mark(Prompt_Mark);         % unsets the region.
        del_region();

        if (strlow(strtrim(user_input)) == "quit"
            and GDB_Trap_Quit)
        {
                gdb_quit();  % Vacate the premises in an orderly fashion :)
        }
        else
        {
                issue_command(user_input, &user_command_hook);
        }
}

static define handle_output (pid, str)
{
        variable pos;

        setbuf ("*gdb*");
        eob ();
        insert (str);
        if (buffer_visible("*gdb*")  %if showing output buffer, update display.
            and GDB_Window_Mode)
        {
                gdb_show_output_buffer();
        }

        if (0 == strncmp (GDB_Output, "(gdb) ", 6))
                GDB_Output = strcat ("\n", GDB_Output);

        GDB_Output += str;

        handle_line_at_a_time(GDB_Output, &generic_line_at_a_time);

        if (Process_Output_Hook != NULL)
        {
                pos = is_substr (GDB_Output, "\n(gdb)");                % Do we have a prompt
                pos += is_substr (GDB_Output, "? (y or n) ");   % or a query?
                if (pos)
                {
                        @Process_Output_Hook(GDB_Output);
                        show_message();
                }
        }
}

public define gdb_mode_prog (Object)
{

        if ((Pid != -1) and bufferp ("*gdb*"))
                error ("gdb-mode can't debug more than 1 process at a time -- can you?");

        Origbuf = whatbuf();
        Source_Buffers = NULL;
        Breakpoints = NULL;
        GDB_Output = "";
        Lineno = NULL;
        Curr_Sourcebuf = NULL;
        Curr_Sourcefile = NULL;
        Process_Output_Hook = NULL;
        Data_Display_Mode =  GDB_Default_Data_Display_Mode;
        Temp_Raw_Mode = 0;
        Msg_Text = "";
        Msg_Priority = -1;

        if (NULL == stat_file (Object))
                verror ("Unable to locate %s", Object);

        pop2buf ("*gdb*");
        local_setkey("gdb_process_buffer_input","^M");

        erase_buffer();
        insert("--> Invoking gdb on " + Object + ":\n");

        Pid = open_process ("gdb","-f", Object, 2);
        set_process (Pid, "signal", "GDB->signal_handler");
        set_process (Pid, "output", "GDB->handle_output");

        wait_for_gdb();

        issue_command("set height 0", NULL);  % No paging, please...
        issue_command("set width 0", NULL);   % ...and no wrapping either.
        issue_command("list 1,0", NULL);     % Have to do this so that info source works
        issue_command("info source", &info_source_hook); % Get filename for source

        if (typeof (Curr_Sourcebuf) != String_Type)
        {
                Curr_Sourcebuf = NULL;
                verror ("No source buffer for gdb");
        }
}

public define gdb_mode_cl ()
{
%	() = evalfile("cmode");
   Object = read_mini("program:", string(__argc), __argv[__argc-1]);
   gdb_mode_prog(Object);
	issue_command("tbreak main", NULL);
	gdb_set_args();
	gdb_run();
}

public define gdb_mode ()
{

        Object = read_with_completion("File to debug: ","","",'f');
        gdb_mode_prog(Object);
}
add_completion("gdb_mode");

#iffalse
public define gdb_mode ()
{

        if ((Pid != -1) and bufferp ("*gdb*"))
                error ("gdb-mode can't debug more than 1 process at a time -- can you?");

        Origbuf = whatbuf();
        Source_Buffers = NULL;
        Breakpoints = NULL;
        GDB_Output = "";
        Lineno = NULL;
        Curr_Sourcebuf = NULL;
        Curr_Sourcefile = NULL;
        Process_Output_Hook = NULL;
        Data_Display_Mode =  GDB_Default_Data_Display_Mode;
        Temp_Raw_Mode = 0;
        Msg_Text = "";
        Msg_Priority = -1;

        Object = read_with_completion("File to debug: ","","",'f');
        if (NULL == stat_file (Object))
                verror ("Unable to locate %s", Object);

        pop2buf ("*gdb*");
        local_setkey("gdb_process_buffer_input","^M");

        erase_buffer();
        insert("--> Invoking gdb on " + Object + ":\n");

        Pid = open_process ("gdb","-f", Object, 2);
        set_process (Pid, "signal", "GDB->signal_handler");
        set_process (Pid, "output", "GDB->handle_output");

        wait_for_gdb();

        issue_command("set height 0", NULL);  % No paging, please...
        issue_command("set width 0", NULL);   % ...and no wrapping either.
        issue_command("list 1,0", NULL);     % Have to do this so that info source works
        issue_command("info source", &info_source_hook); % Get filename for source

        if (typeof (Curr_Sourcebuf) != String_Type)
        {
                Curr_Sourcebuf = NULL;
                verror ("No source buffer for gdb");
        }
}

#endif
